﻿<#
.SYNOPSIS
	Uploads multiple LPE firmware updates to all available pools.

.DESCRIPTION
	Uploades Lync Phone Edition device firmware to all pools. It supports uploading firmware for multiple device models and multiple pools

.NOTES
  Version      				  : 2.1 - See changelog at https://www.ucunleashed.com/1245
  Wish List						  :
  Rights Required			  : Local admin on server
  										  : RTCUniversalServerAdmins or equivilent group membership that passes
  										  : Get-CsAdminRole | Where-Object {$_.Cmdlets –match "Import-CsDeviceUpdate"}
  Lync/Skype4B Version  : 2010 (tested on CU 6); 2013 (tested on CU 10)
  Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
  Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
  Donations             : https://www.paypal.me/PatRichard
  Dedicated Blog			  : https://www.ucunleashed.com/1217
  Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                          provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                          without limitation, any implied warranties of merchantability or of fitness for a particular
                          purpose. The entire risk arising out of the use or performance of the sample scripts and
                          documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                          (including, without limitation, damages for loss of business profits, business interruption,
                          loss of business information, or other pecuniary loss) arising out of the use of or inability
                          to use the script or documentation. Neither this script, nor any part of it other than those
                          parts that are explicitly copied from others, may be republished without author(s) express written 
                          permission.
	Acknowledgements		  : Kevin Peters for some original code examples
											  : Tom Arbuthnot for finding a bug when running the script from ISE
	Assumptions					  : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
	Limitations					  :
	Known Issues				  :	None yet, but I'm sure you'll find some!

.LINK
  https://www.ucunleashed.com/1217

.EXAMPLE
	.\New-CsFirmware.ps1

	Details
	=======
	Will prompt for a parent folder containing any firmware files, including in child folders, and will upload them to all pools that it finds.

.EXAMPLE
	.\New-CsFirmware.ps1 -FilesPath [path to parent folder]

	Details
	=======
	Will upload firmware update files within the parent folder and any child folder to all pools that it finds.

.INPUTS
	None. You cannot pipe objects to this script.
#>
#Requires -Version 2.0

[CmdletBinding(SupportsShouldProcess)]
param(
	# path of folder containing new firmware files
	[Parameter(ValueFromPipelineByPropertyName)]
	[string] $FilesPath,

	# If specified, will download all firmware before importing
	[Parameter(ValueFromPipelineByPropertyName)]
	[switch] $Download,

	# Location of logs folder
	[Parameter(ValueFromPipelineByPropertyName)]
	[ValidatePattern('^(?:[a-zA-Z]\:|\\\\[\w\.]+\\[\w.]+)\\(?:[\w]+\\)*\w([\w.])+$')]
	[string] $TargetFolder = "$env:SystemDrive\_Install",

	# FQDN of a single pool to upload firmware files to. If not specified, files are uploaded to all pools
	[Parameter(ValueFromPipelineByPropertyName)]
	[string] $PoolFqdn
)
[string] $ScriptVersion = "2.00"
[int] $Article = 1245
[string] $LogPath = "$TargetFolder\FirmwareLogs\$env:ComputerName" + " {0:yyyy-MM-dd hh-mmtt}.log" -f (Get-Date)
[string] $LogDivider = "------------------------------"

#region functions
Write-Verbose "Defining functions"
function Get-Folder {
	<#
	.SYNOPSIS
		Displays a folder dialog box, and returns the selected folder. Also allows for a new folder to be created.

	.DESCRIPTION
		Displays a folder dialog box, and returns the selected folder. Also allows for a new folder to be created.

	.NOTES
	  Version							  : 1.3
		Wish list						  : Better error trapping
	  Rights Required			  : None
	  Sched Task Required	  : No
	  Lync Server Version	  : N/A
	  Author/Copyright		  : © Pat Richard, Skype for Business MVP - All Rights Reserved
	  Email/Blog/Twitter	  : pat@innervation.com 	https://www.ucunleashed.com @patrichard
	  Donations             : https://www.paypal.me/PatRichard
	  Dedicated Post			  :
	  Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                            provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                            without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and
                            documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                            (including, without limitation, damages for loss of business profits, business interruption,
                            loss of business information, or other pecuniary loss) arising out of the use of or inability
                            to use the script or documentation. Neither this script, nor any part of it other than those
                            parts that are explicitly copied from others, may be republished without author(s) express written 
                            permission.
	  Acknowledgements 		  :
	  Assumptions					  : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
	  Limitations					  :
	  Known issues				  :

	.LINK


	.EXAMPLE
	  Get-Folder

	  Description
		-----------
		Displays a folder dialog box, and returns the selected folder. Also allows for a new folder to be created.

	.INPUTS
		None. You cannot pipe objects to this script.
	#>
	[CmdletBinding(SupportsShouldProcess, SupportsPaging)]
	param(
		# Message that is displayed to the user when the browse folder dialong box opens
		[Parameter(ValueFromPipelineByPropertyName)]
		[ValidateNotNullOrEmpty()]
		[string] $Message = "Select the desired folder",

		# Starting location for the dialog box. Options are:
		# http://msdn.microsoft.com/en-us/library/windows/desktop/bb774096
		<#
		  ssfALTSTARTUP        = 0x1d,
		  ssfAPPDATA           = 0x1a,
		  ssfBITBUCKET         = 0x0a,
		  ssfCOMMONALTSTARTUP  = 0x1e,
		  ssfCOMMONAPPDATA     = 0x23,
		  ssfCOMMONDESKTOPDIR  = 0x19,
		  ssfCOMMONFAVORITES   = 0x1f,
		  ssfCOMMONPROGRAMS    = 0x17,
		  ssfCOMMONSTARTMENU   = 0x16,
		  ssfCOMMONSTARTUP     = 0x18,
		  ssfCONTROLS          = 0x03,
		  ssfCOOKIES           = 0x21,
		  ssfDESKTOP           = 0x00,
		  ssfDESKTOPDIRECTORY  = 0x10,
		  ssfDRIVES            = 0x11,
		  ssfFAVORITES         = 0x06,
		  ssfFONTS             = 0x14,
		  ssfHISTORY           = 0x22,
		  ssfINTERNETCACHE     = 0x20,
		  ssfLOCALAPPDATA      = 0x1c,
		  ssfMYPICTURES        = 0x27,
		  ssfNETHOOD           = 0x13,
		  ssfNETWORK           = 0x12,
		  ssfPERSONAL          = 0x05,
		  ssfPRINTERS          = 0x04,
		  ssfPRINTHOOD         = 0x1b,
		  ssfPROFILE           = 0x28,
		  ssfPROGRAMFILES      = 0x26,
		  ssfPROGRAMFILESx86   = 0x30,
		  ssfPROGRAMS          = 0x02,
		  ssfRECENT            = 0x08,
		  ssfSENDTO            = 0x09,
		  ssfSTARTMENU         = 0x0b,
		  ssfSTARTUP           = 0x07,
		  ssfSYSTEM            = 0x25,
		  ssfSYSTEMx86         = 0x29,
		  ssfTEMPLATES         = 0x15,
		  ssfWINDOWS           = 0x24
		#>
		[Parameter]
		[int] $path = 0x00
	)
  [Object] $FolderObject = New-Object -ComObject Shell.Application
  $folder = $FolderObject.BrowseForFolder(0, $message, 0, $path)
  if ($folder -ne $null) {
		return $folder.self.Path
  } else {
  	Write-Output "No folder specified"
  }
} # end function Get-Folder

function Get-UpdateInfo {
<#
.SYNOPSIS
  Queries an online XML source for version information to determine if a new version of the script is available.

.DESCRIPTION
  Queries an online XML source for version information to determine if a new version of the script is available.

.NOTES
  Version               : 1.2 - See changelog at https://www.ucunleashed.com/3168 for fixes & changes introduced with each version
  Wish list             : Better error trapping
  Rights Required       : N/A
  Sched Task Required   : No
  Lync/Skype4B Version  : N/A
  Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
  Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
  Donations             : https://www.paypal.me/PatRichard
  Dedicated Post        : https://www.ucunleashed.com/3168
  Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                          provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                          without limitation, any implied warranties of merchantability or of fitness for a particular
                          purpose. The entire risk arising out of the use or performance of the sample scripts and
                          documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                          (including, without limitation, damages for loss of business profits, business interruption,
                          loss of business information, or other pecuniary loss) arising out of the use of or inability
                          to use the script or documentation. Neither this script, nor any part of it other than those
                          parts that are explicitly copied from others, may be republished without author(s) express written 
                          permission.
  Acknowledgements      : Reading XML files 
                          http://stackoverflow.com/questions/18509358/how-to-read-xml-in-powershell
                          http://stackoverflow.com/questions/20433932/determine-xml-node-exists
  Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
  Limitations           : 
  Known issues          : 

.LINK
  https://www.ucunleashed.com/3168

.EXAMPLE
  Get-UpdateInfo -Article 123

  Description
  -----------
  Runs function to check for updates to article/script 123.

.INPUTS
  None. You cannot pipe objects to this script.
#>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging)]
  param (
    
    [ValidateNotNullOrEmpty()]
    [string] $article
  )
  [bool] $HasInternetAccess = ([Activator]::CreateInstance([Type]::GetTypeFromCLSID([Guid]'{DCB00C01-570F-4A9B-8D69-199FDBA5723B}')).IsConnectedToInternet)
  if ($HasInternetAccess){
    [xml] $xml = (New-Object System.Net.WebClient).DownloadString("https://www.ucunleashed.com/downloads/version.xml")
    [string] $Ga = ($xml.catalog.article | Where-Object {$_.id -eq $article}).version
    if (($xml.catalog.article | Where-Object {$_.id -eq $article}).versionInfo){
      [string] $changelog = "This version includes: " + ($xml.catalog.article | Where-Object {$_.id -eq $article}).versionInfo
    }
    if ($Ga -gt $ScriptVersion){
      Write-Log -Level Warn -Message "Outdated version. Version $Ga is latest version. Prompting user to upgrade." -NoConsole
      if ((New-Popup -Message "Version $Ga is now available. $changelog `n`nWould you like to download it?" -Title "New version available!" -Buttons YesNo -Icon Question) -eq 6){
        Write-Log -Message "User elected to download latest version." -Indent 1 -NoConsole
        Write-Log -Message "Opening Internet Explorer and going to https://www.ucunleashed.com/$article#downloads." -NoConsole -Indent 1
        Start-Process "https://www.ucunleashed.com/$article#downloads"
        Write-Log -Level Warn -Message "Script is exiting. Please start the new version of the script once you've downloaded it." -NoLog
        Exit
      }else{
        Write-Log -Level Warn -Message "User elected NOT to download latest version." -Indent 1 -NoConsole
      }
    }elseif ($Ga -eq $ScriptVersion){
      Write-Log -Message "Current version. Version $Ga is latest version." -NoConsole
    }elseif (-Not ($Ga)) {
      # Write-Log -Level Warn -Message "Possibly regressed version. Version $Ga is latest version." -NoConsole
      Write-Log -Level Warn -Message "No version info available" -NoConsole
    }
  }else{
    Write-Log -Level Warn -Message "Unable to check online for update info." -NoConsole
  }
} # end function function Get-UpdateInfo

function Set-ModuleStatus {
	<#
	.SYNOPSIS
		Imports a specified PowerShell module, with error checking.

	.DESCRIPTION
		Imports a specified PowerShell module, with error checking.

	.NOTES
	  Version							  : 1.6
		Wish list						  : Better error trapping
	  Rights Required			  : None
	  Sched Task Required	  : No
	  Lync Server Version	  : N/A
	  Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
    Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
    Donations             : https://www.paypal.me/PatRichard
	  Dedicated Post			  :
    Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                            provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                            without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and
                            documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                            (including, without limitation, damages for loss of business profits, business interruption,
                            loss of business information, or other pecuniary loss) arising out of the use of or inability
                            to use the script or documentation. Neither this script, nor any part of it other than those
                            parts that are explicitly copied from others, may be republished without author(s) express written 
                            permission.
	  Acknowledgements 		  :
	  Assumptions					  : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
	  Limitations					  :
	  Known issues				  :

	.LINK


	.EXAMPLE
	  .\

	  Description
		-----------


	.INPUTS
		None. You cannot pipe objects to this script.
	#>
	[CmdletBinding(SupportsShouldProcess, SupportsPaging)]
	param	(
		[parameter(ValueFromPipeline, ValueFromPipelineByPropertyName, Mandatory, HelpMessage = "No module name specified!")]
		[ValidateNotNullOrEmpty()]
		[string] $name
	)
	PROCESS{
		# Executes once for each pipeline object
		# the $_ variable represents the current input object
		if (-not(Get-Module -name "$name")) {
			if (Get-Module -ListAvailable | Where-Object Name -eq "$name") {
				Import-Module -Name "$name"
				# module was imported
				# return $true
			} else {
				# module was not available
				# return $false
			}
		} else {
			# Write-Output "$_ module already imported"
			# return $true
		}
	} # end PROCESS
} # end function Set-ModuleStatus

function Test-IsProxyEnabled {
	<#
	.SYNOPSIS
		Determines (the best it can) if a proxy is in place for Internet access.

	.DESCRIPTION
		Determines (the best it can) if a proxy is in place for Internet access. If so, other functions that download files can better handle downloads.

	.NOTES
	  Version							  : 1.2
		Wish list						  : Better error trapping
	  Rights Required			  : Local administrator on server
	  Sched Task Required	  : No
	  Lync Server Version	  : N/A
	  Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
    Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
    Donations             : https://www.paypal.me/PatRichard
	  Dedicated Post			  :
    Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                            provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                            without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and
                            documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                            (including, without limitation, damages for loss of business profits, business interruption,
                            loss of business information, or other pecuniary loss) arising out of the use of or inability
                            to use the script or documentation. Neither this script, nor any part of it other than those
                            parts that are explicitly copied from others, may be republished without author(s) express written 
                            permission.
	  Acknowledgements 		  :
	  Assumptions					  : ExecutionPolicy of AllSigned (recommended), RemoteSigned or Unrestricted (not recommended)
	  Limitations					  :
	  Known issues				  :

	.LINK


	.EXAMPLE
	  .\

	  Description
		-----------


	.INPUTS
		None. You cannot pipe objects to this script.
	#>
	[CmdletBinding(SupportsShouldProcess, SupportsPaging)]
	param ()
	Write-Log -Message "Checking if a proxy is enabled" -NoConsole -Indent 1
	if ((Get-ItemProperty "HKCU:\Software\Microsoft\Windows\CurrentVersion\Internet Settings" -Name ProxyEnable -ErrorAction SilentlyContinue).ProxyEnable){
		Write-Log -Message "A proxy is required" -NoConsole -Indent 2
		[string] $ProxyServer = (Get-ItemProperty "HKCU:\Software\Microsoft\Windows\CurrentVersion\Internet Settings" -name ProxyServer).ProxyServer
		Write-Log -Message "Configured proxy: `"$ProxyServer`"" -NoConsole -Indent 2
		# http://www.powershellmagazine.com/2013/02/06/pstip-validate-if-proxy-credentials-are-required/
		if (-not($PSDefaultParameterValues.ContainsKey("Start-BitsTransfer:ProxyAuthentication"))){
			Write-Log -Message "Prompting for credentials" -NoConsole -Indent 2
			$ProxyCredentials = Get-Credential -Message "Enter Proxy authentication credentials for $ProxyServer" -UserName "${env:USERDOMAIN}\${env:UserName}"
    	if ($ProxyCredentials){
    		Write-Log -Message "Credentials entered" -NoConsole -Indent 2
    		Write-Log -Message "Adding default values for Start-BitsTransfer" -NoConsole -Indent 2
    		Write-Log -Message "Adding default value for ProxyAuthentication - `"Basic`"" -NoConsole -Indent 2
    		$PSDefaultParameterValues.Add("Start-BitsTransfer:ProxyAuthentication","Basic")
    		$ProxyUserName = $ProxyCredentials.username
    		Write-Log -Message "Adding default value for ProxyCredential - `"$ProxyUserName`"" -NoConsole -Indent 2
    		$PSDefaultParameterValues.Add("Start-BitsTransfer:ProxyCredential",$ProxyCredentials)
    		# Write-Log -Message "Adding default value for ProxyList - `"$ProxyServer`"" -NoConsole -Indent 1
    		# $PSDefaultParameterValues.Add("Start-BitsTransfer:ProxyList",$ProxyServer)
    		# Write-Log -Message "Adding default value for ProxyUsage - `"AutoDetect`"" -NoConsole -Indent 1
    		# $PSDefaultParameterValues.Add("Start-BitsTransfer:ProxyUsage","AutoDetect")
			} else {
				Write-Log -Level error -message "Credentials NOT entered. Following commands will likely fail" -NoConsole -Indent 2
			}
    }
	} else {
		Write-Log -Message "Proxy is not enabled" -NoConsole -Indent 2
	}
} # end function Test-IsProxyEnabled

function New-FileDownload {
  <#
  .SYNOPSIS
    Downloads a file from a specified URL.

  .DESCRIPTION
    Downloads a file from a specified URL. First, it verifies the file does not exist locally. Then ensures Internet access is available, then downloads the file.

  .NOTES
    Version               : 1.5
    Wish list             : Better error trapping
    Rights Required       : Local administrator on server
    Sched Task Required   : No
    Lync/Skype4B Version  : N/A
    Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
    Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
    Donations             : https://www.paypal.me/PatRichard
    Dedicated Post        :
    Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                            provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                            without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and
                            documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                            (including, without limitation, damages for loss of business profits, business interruption,
                            loss of business information, or other pecuniary loss) arising out of the use of or inability
                            to use the script or documentation. Neither this script, nor any part of it other than those
                            parts that are explicitly copied from others, may be republished without author(s) express written 
                            permission.
    Acknowledgements      :
    Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
    Limitations           :
    Known issues          : None yet, but I'm sure you'll find some!

  .LINK

  .EXAMPLE
    .\

    Description
    -----------


  .INPUTS
    None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging)]
  param(
    # Complete path and file name (URL) to the file to be downloaded
    [parameter(ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [string] $SourceFile,

    # The folder where the downloaded file should be placed. If not defined, it defaults to $TargetFolder.
    [parameter(ValueFromPipelineByPropertyName)]
    [string] $DestFolder,

    # The file name the downloaded file should be changed to. If not defined, file maintains it's original name.
    [parameter(ValueFromPipelineByPropertyName)]
    [string] $DestFile,

    # Whether to download even if a local copy exists. This is useful for files that are updated often and need to be redownloaded.
    [parameter(ValueFromPipelineByPropertyName)]
    [switch] $IgnoreLocalCopy
  ) # end of param block
  [bool] $HasInternetAccess = ([Activator]::CreateInstance([Type]::GetTypeFromCLSID([Guid]'{DCB00C01-570F-4A9B-8D69-199FDBA5723B}')).IsConnectedToInternet)
  Test-IsProxyEnabled
  # I should switch this to using a param block and pipelining from property name - just for consistency

  if (-Not($DestFolder)){
    $DestFolder = $TargetFolder
  }
  Set-ModuleStatus -Name BitsTransfer
  if (-Not($DestFile)){
    # [string] $DestFile = ($SourceFile | Split-Path -leaf)
    Add-Type -AssemblyName System.Web
    [string] $DestFile = [System.Web.HttpUtility]::UrlDecode(($SourceFile | Split-Path -leaf))
    [Reflection.Assembly]::LoadWithPartialName("System.Web") | Out-Null
    $DestFile = [System.Web.HttpUtility]::UrlDecode($DestFile)
  }
  if (Test-Path -Path $DestFolder){
    Write-Log -Message "Target folder `"$DestFolder`" exists - no need to create" -NoConsole -Indent 1
  } else {
    Write-Log -Message "Folder `"$DestFolder`" does not exist, creating..." -NoConsole
    New-Item $DestFolder -type Directory | Out-Null
    Write-Log -Message "Done!" -NoConsole -Indent 1
  }
  if ((Test-Path -Path "$DestFolder\$DestFile") -and (-Not $IgnoreLocalCopy)){
    if ($(((Get-Item $DestFolder\$DestFile).VersionInfo).FileVersion)){
      Write-Log -Message "File `"$DestFile`" version $(((Get-Item $DestFolder\$DestFile).VersionInfo).FileVersion) exists locally - no need to download" -NoConsole
    }else{
      Write-Log -Message "File `"$DestFile`" exists locally - no need to download" -NoConsole
    }
  } else {
    if ($HasInternetAccess){
      Write-Log -Message "Internet access available" -NoConsole -Indent 1
      if (-Not $IgnoreLocalCopy){
        Write-Log -Message "File `"$DestFile`" does not exist in `"$DestFolder`"" -NoConsole -Indent 1
      } else {
        Write-Log -Message "Forcing download of `"$DestFile`" to `"$DestFolder`"" -NoConsole
      }
      Write-Log -Message "Downloading `"$SourceFile`" to `"$DestFolder`"" -NoConsole -Indent 1
      ########################################################################################################
      # NOTE: Default parameters may have been changed due to proxy settings. See Test-IsProxyEnabled function
      ########################################################################################################
      # determine file size before downloading
      <#
      $clnt = New-Object System.Net.WebClient
      $clnt.OpenRead($SourceFile) | Out-Null
      $dlfilesize = [int] $($clnt.ResponseHeaders["Content-Length"]/1mb)
      # $dlfilesize
      $clnt.Dispose()
      #>
      if ($dlfilesize){
        Start-BitsTransfer -Source "$SourceFile" -Destination "$DestFolder\$DestFile" -Description "Downloading $DestFile ($dlfilesize MB)" -ErrorAction SilentlyContinue
      }else{
        Start-BitsTransfer -Source "$SourceFile" -Destination "$DestFolder\$DestFile" -Description "Downloading $DestFile" -ErrorAction SilentlyContinue
      }

      if (Test-Path -Path $DestFolder\$DestFile){
        if ($(((Get-Item $DestFolder\$DestFile).VersionInfo).FileVersion)){
          Write-Log -Message "Successfully downloaded $DestFolder\$DestFile version $(((Get-Item $DestFolder\$DestFile).VersionInfo).FileVersion)" -NoConsole -Indent 1
        }else{
          Write-Log -Message "Successfully downloaded $DestFolder\$DestFile" -NoConsole -Indent 1
        }
      } else {
        Write-Log -Level error -message "Failed! File `"$DestFile`" not downloaded!" -Indent 1
        Write-Log -Message "Prompting user to abort/retry/ignore" -NoConsole
        switch (New-Popup -message "Download failure has occurred for `"$DestFile`". What would you like to do?" -Title "Download error!" -Buttons "AbortRetryIgnore" -Icon Exclamation){
          3{ # abort
            Write-Log -Message "User has chosen to abort script" -NoConsole -Indent 1
            Stop-Script
            exit
          }
          4{ # retry
            Write-Log -Message "User has chosen to retry download of `"$DestFile`"" -NoConsole -Indent 1
            Write-Log -Message "Building retry expression" -NoConsole -Indent 2
            if ($IgnoreLocalCopy){
              $DownloadRetry += " -IgnoreLocalCopy"
            }
            if ($DestFile){
              $DownloadRetry += " -DestFile `"$Destfile`""
            }
            if ($DestFolder){
              $DownloadRetry += " -DestFolder `"$DestFolder`""
            }
            $DownloadRetryExp = "New-FileDownload -SourceFile `"$SourceFile`""+ $DownloadRetry
            Write-Log -Message "Retry expression is $DownloadRetryExp" -NoConsole -Indent 2
            Invoke-Expression $DownloadRetryExp
          }
          5{ # ignore
            Write-Log -Message "User has chosen to ignore" -NoConsole -Indent 1
          }
        }
      }
    } else {
      Write-Log -Level warn -Message "Internet access not detected."
      Write-Log -Level warn -Message "This can be because there is no Internet connection,"
      Write-Log -Level warn -Message "there is no DNS resolution,"
      Write-Log -Level warn -Message "or a proxy is in place. Please resolve and try again."
      Write-Log -Level warn -Message "Alternatively, you can manually download the file ($SourceFile)"
      Write-Log -Level warn -Message "and place it in $DestFolder and try again."
    }
  }
} # end function New-FileDownload

function Write-Log {
  <#
  .SYNOPSIS
    Extensive function to write data to either the console screen, a log file, and/or a Windows event log.

  .DESCRIPTION
    Extensive function to write data to either the console screen, a log file, and/or a Windows event log. Data can be written as info, warning, error, and includes indentation, time stamps, etc.

  .NOTES
    Version               : 2.9
    Wish list             : Better error trapping
    Rights Required       : Local administrator on server
    Sched Task Required   : No
    Lync/Skype4B Version  : N/A
    Author/Copyright      : © Pat Richard, Office Servers and Services (Skype for Business) MVP - All Rights Reserved
    Email/Blog/Twitter    : pat@innervation.com  https://ucunleashed.com  @patrichard
    Donations             : https://www.paypal.me/PatRichard
    Dedicated Post        :
    Disclaimer            : You running this script means you won't blame author(s) if this breaks your stuff. This script is
                            provided AS IS without warranty of any kind. Author(s) disclaim all implied warranties including,
                            without limitation, any implied warranties of merchantability or of fitness for a particular
                            purpose. The entire risk arising out of the use or performance of the sample scripts and
                            documentation remains with you. In no event shall author(s) be liable for any damages whatsoever
                            (including, without limitation, damages for loss of business profits, business interruption,
                            loss of business information, or other pecuniary loss) arising out of the use of or inability
                            to use the script or documentation. Neither this script, nor any part of it other than those
                            parts that are explicitly copied from others, may be republished without author(s) express written 
                            permission.
    Acknowledgements      : Test for log names and sources
                            http://powershell.com/cs/blogs/tips/archive/2013/06/10/testing-event-log-names-and-sources.aspx
    Assumptions           : ExecutionPolicy of AllSigned (recommended), RemoteSigned, or Unrestricted (not recommended)
    Limitations           :
    Known issues          : None yet, but I'm sure you'll find some!

  .LINK

  .EXAMPLE
    .\

    Description
    -----------


  .INPUTS
    None. You cannot pipe objects to this script.
  #>
  [CmdletBinding(SupportsShouldProcess, SupportsPaging)]
  param(
    # The type of message to be logged. Alias is 'type'.
    [Parameter(Position = 0, ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateSet("Error", "Warn", "Info")]
    [ValidateNotNullOrEmpty()]
    [string] $level = "Info",

    # The message to be logged.
    [Parameter(Position = 1, ValueFromPipeline, ValueFromPipelineByPropertyName, Mandatory, HelpMessage = "No message specified.")]
    [ValidateNotNullOrEmpty()]
    [string] $Message,

    # Specifies that $message should not the sent to the log file.
    [Parameter(Position = 2, ValueFromPipelineByPropertyName)]
    [switch] $NoLog,

    # Specifies to not display the message to the console.
    [Parameter(Position = 3, ValueFromPipelineByPropertyName)]
    [switch] $NoConsole,

    # The number of spaces to indent the message in the log file.
    [Parameter(Position = 4, ValueFromPipelineByPropertyName)]
    [ValidateRange(1,30)]
    [ValidateNotNullOrEmpty()]
    [Int16] $Indent = 0,

    # Specifies what color the text should be be displayed on the console. Ignored when switch 'NoConsoleOut' is specified.
    [Parameter(Position = 5, ValueFromPipelineByPropertyName)]
    [ValidateSet("Black", "DarkMagenta", "DarkRed", "DarkBlue", "DarkGreen", "DarkCyan", "DarkYellow", "Red", "Blue", "Green", "Cyan", "Magenta", "Yellow", "DarkGray", "Gray", "White")]
    [ValidateNotNullOrEmpty()]
    [String] $ConsoleForeground = 'White',

    # Existing log file is deleted when this is specified. Alias is 'Overwrite'.
    [Parameter(Position = 6, ValueFromPipelineByPropertyName)]
    [Switch] $Clobber,

    # The name of the system event log, e.g. 'Application'. Note that writing to the system event log requires elevated permissions.
    [Parameter(Position = 7, ValueFromPipelineByPropertyName)]
    [ValidateSet("Application","System","Security","Lync Server","Microsoft Office Web Apps")]
    [ValidateNotNullOrEmpty()]
    [String] $EventLogName,

    # The name to appear as the source attribute for the system event log entry. This is ignored unless 'EventLogName' is specified.
    [Parameter(Position = 8, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [String] $EventSource = $($MyInvocation.ScriptName).Name,

    # The ID to appear as the event ID attribute for the system event log entry. This is ignored unless 'EventLogName' is specified.
    [Parameter(Position = 9, ValueFromPipelineByPropertyName)]
    [ValidateRange(1,65535)]
    [ValidateNotNullOrEmpty()]
    [Int32] $EventID = 1,

    # The text encoding for the log file. Default is ASCII.
    [Parameter(Position = 10, ValueFromPipelineByPropertyName)]
    [ValidateSet("Unicode","Byte","BigEndianUnicode","UTF8","UTF7","UTF32","ASCII","Default","OEM")]
    [ValidateNotNullOrEmpty()]
    [String] $LogEncoding = "ASCII"
  ) # end of param block
  try {
    [string]$LogFolder = Split-Path $LogPath -Parent
    if (-Not(Test-Path -Path $LogFolder)){New-Item $LogFolder -type Directory | Out-Null}
    $msg = "{0} : {1} : {2}{3}" -f (Get-Date -Format "yyyy-MM-dd HH:mm:ss"), $Level.ToUpper(), ("  " * $Indent), $Message
    if (-Not($NoConsole)){
      switch ($level) {
        "Error" {$Host.UI.WriteErrorLine("$Message")}
        "Warn" {Write-Warning $Message}
        "Info" {Write-Host $Message -ForegroundColor $ConsoleForeground}
      }
    }
    if (-Not($NoLog)){
      if ($Clobber) {
        $msg | Out-File -FilePath $LogPath -Encoding $LogEncoding -Force
      } else {
        $msg | Out-File -FilePath $LogPath -Encoding $LogEncoding -Append
      }
    }

    # http://social.technet.microsoft.com/Forums/en-US/winserverpowershell/thread/e172f039-ce88-4c9f-b19a-0dd6dc568fa0/

    if ($EventLogName) {
      if (-Not $EventSource) {
        [string] $EventSource = $([IO.FileInfo] $MyInvocation.ScriptName).Name
      }

      if(-Not [Diagnostics.EventLog]::SourceExists($EventSource)) {
        [Diagnostics.EventLog]::CreateEventSource($EventSource, $EventLogName)
      }

      switch ($Level) {
        "Error" {$EntryType = "Error"}
        "Warn"  {$EntryType = "Warning"}
        "Info"  {$EntryType = "Information"}
        Default  {$EntryType = "Information"}
      }
      Write-EventLog -LogName $EventLogName -Source $EventSource -EventId 1 -EntryType $EntryType -Message $Message
    }
    $msg = ""
  } # end try
  catch {
    Throw "Failed to create log entry in: '$LogPath'. The error was: '$_'."
  } # end catch
} # end function Write-Log
#endregion functions

Get-UpdateInfo -Article $Article

Write-Log -Message "Importing Lync module" -NoConsole
Set-ModuleStatus Lync

if ($Download){
	if (-not(Test-Path "$TargetFolder\7z.exe")){
		Write-Log -Message "Downloading 7z files for use in extracting files from LPE firmware" -NoConsole
		New-FileDownload -SourceFile "https://www.ucunleashed.com/downloads/1217/7z.dll_" -DestFolder $TargetFolder -DestFile "7z.dll"
		New-FileDownload -SourceFile "https://www.ucunleashed.com/downloads/1217/7z.exe_" -DestFolder $TargetFolder -DestFile "7z.exe"
		Write-Log -Message "Done!" -Indent 1 -NoConsole
	}else{
		Write-Log -Message "7z files already exist" -NoConsole -Indent 1
	}

	Push-Location
	$today = Get-Date -Format yyyy-MM-dd
	[string] $LPEfolder = "$TargetFolder\Lync Phone Edition updates\$today"
	[string] $ExtractCodeBlock = "$TargetFolder\7z.exe e ucupdates.exe ucupdates.cab -aoa"

	New-FileDownload -SourceFile "http://download.microsoft.com/download/F/8/8/F88C7D24-FCCB-41E4-A0A0-FCA813C3FDD9/UCUpdates.exe" -DestFolder "$LPEfolder\Microsoft Lync Phone Edition for Polycom CX700 and LG-Nortel IP Phone 8540"
	Set-Location "$LPEfolder\Microsoft Lync Phone Edition for Polycom CX700 and LG-Nortel IP Phone 8540"
	Write-Log -Message "Extracting ucupdates.cab from $LPEfolder\Microsoft Lync Phone Edition for Polycom CX700 and LG-Nortel IP Phone 8540\ucupdates.exe" -NoConsole
	Invoke-Expression $ExtractCodeBlock | Out-Null
	Write-Log -Message "Done!" -Indent 1 -NoConsole

	New-FileDownload -SourceFile "http://download.microsoft.com/download/2/8/D/28DA8A01-C9EC-4F3E-A55A-FC5F46DBF0A1/UCUpdates.exe" -DestFolder "$LPEfolder\Microsoft Lync Phone Edition for Polycom CX500, Polycom CX600 and Polycom CX3000"
	Set-Location "$LPEfolder\Microsoft Lync Phone Edition for Polycom CX500, Polycom CX600 and Polycom CX3000"
	Write-Log -Message "Extracting ucupdates.cab from $LPEfolder\Microsoft Lync Phone Edition for Polycom CX500, Polycom CX600 and Polycom CX3000\ucupdates.exe" -NoConsole
	Invoke-Expression $ExtractCodeBlock | Out-Null
	Write-Log -Message "Done!" -Indent 1 -NoConsole

	New-FileDownload -SourceFile "http://download.microsoft.com/download/C/A/E/CAEB181B-7A20-49C6-BDDE-D676EA8F02A1/UCUpdates.exe" -DestFolder "$LPEfolder\Microsoft Lync Phone Edition for Aastra 6721ip and Aastra 6725ip"
	Set-Location "$LPEfolder\Microsoft Lync Phone Edition for Aastra 6721ip and Aastra 6725ip"
	Write-Log -Message "Extracting ucupdates.cab from $LPEfolder\Microsoft Lync Phone Edition for Aastra 6721ip and Aastra 6725ip\ucupdates.exe" -NoConsole
	Invoke-Expression $ExtractCodeBlock | Out-Null
	Write-Log -Message "Done!" -Indent 1 -NoConsole

	New-FileDownload -SourceFile "http://download.microsoft.com/download/8/8/C/88CDF8C2-30D2-4189-8124-2E57D0078EE5/UCUpdates.exe" -DestFolder "$LPEfolder\Microsoft Lync Phone Edition for HP 4110 and HP 4120"
	Set-Location "$LPEfolder\Microsoft Lync Phone Edition for HP 4110 and HP 4120"
	Write-Log -Message "Extracting ucupdates.cab from $LPEfolder\Microsoft Lync Phone Edition for HP 4110 and HP 4120\ucupdates.exe" -NoConsole
	Invoke-Expression $ExtractCodeBlock | Out-Null
	Write-Log -Message "Done!" -Indent 1 -NoConsole

	Pop-Location
	$FilesPath = $LPEfolder
}

Write-Log -Message "Getting pool list of Lync pools, and ignoring OCS pools" -NoConsole
if (-not ($PoolFqdn)){
	$FEPools = Get-CSService -UserServer | Where-Object {$_.Version -ge 5}
}else{
	$FEPools = Get-CSService -UserServer -PoolFqdn $PoolFqdn
}
if (-not ($FilesPath)){
	Write-Log -Message "Path not specified. Prompting for folder." -NoConsole
	$FilesPath = (Get-Folder "Select the folder that contains the firmware files")
}
Write-Log -Message "Getting list of firmware update files" -NoConsole
[object] $files = Get-ChildItem -Path $FilesPath -Recurse | Where-Object {$_.name -match ".cab$"}
[int] $filesfound = $files.count
Write-Log -Message "$filesfound files found" -NoConsole -Indent 1
if ($Files -and $FEPools){
	ForEach ($File in $Files){
		ForEach ($FEPool in $FEPools) {
			Write-Log -Message "Importing $($file.versioninfo.filename) to $FEPool" -NoConsole
			[string] $pool = "service:WebServer:"+$FEPool.PoolFQDN
			Import-CsDeviceUpdate -Identity $pool -FileName $file.versioninfo.filename -Verbose
		}
	}
}
# SIG # Begin signature block
# MIIcdgYJKoZIhvcNAQcCoIIcZzCCHGMCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUE0NSVDRlu8iQW0EYGbfgT+YJ
# jqqgghelMIIFLjCCBBagAwIBAgIQA6cFj2F2VqaerCviGAu9jTANBgkqhkiG9w0B
# AQsFADByMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYD
# VQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYDVQQDEyhEaWdpQ2VydCBTSEEyIEFz
# c3VyZWQgSUQgQ29kZSBTaWduaW5nIENBMB4XDTE2MTAxMjAwMDAwMFoXDTE3MTIy
# MTEyMDAwMFowazELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAk1JMRkwFwYDVQQHExBT
# dGVybGluZyBIZWlnaHRzMRkwFwYDVQQKExBJbm5lcnZhdGlvbiwgTExDMRkwFwYD
# VQQDExBJbm5lcnZhdGlvbiwgTExDMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
# CgKCAQEAvVNbtzSVPbpRQfYrhU0lPzCnmYKS2DFpgrL8gI8qUsyojdQJRjF8q5YQ
# uIi92Uwl/hxWcvCf0dU/EbzRZudOr8tsbkLA6x1R7O5Fsd6Q2fwMN4YmhFDKUcxf
# Gjy1byOdEhSQZqUc4xp6T+prcIi1j4EDSp5R0TQSRiKUK5HoqanhRyGfqJjkEOep
# KB9soxNVkB+jTXsIjX1AeuUJZS/m+1rzWf75laoid+rna7KQtJO+DkVNSa96tTH2
# 1Fs/C78UzWBLA5I2gQAgEA3alkMzB7rIr34Grt251O+yMSyCcvjb70AVcKohoS7n
# 2Zk522EK3jJw3F8euSWMbXNegJtPjwIDAQABo4IBxTCCAcEwHwYDVR0jBBgwFoAU
# WsS5eyoKo6XqcQPAYPkt9mV1DlgwHQYDVR0OBBYEFE+s6yTjN7ocNskZ2gxbJ3Ul
# xBo6MA4GA1UdDwEB/wQEAwIHgDATBgNVHSUEDDAKBggrBgEFBQcDAzB3BgNVHR8E
# cDBuMDWgM6Axhi9odHRwOi8vY3JsMy5kaWdpY2VydC5jb20vc2hhMi1hc3N1cmVk
# LWNzLWcxLmNybDA1oDOgMYYvaHR0cDovL2NybDQuZGlnaWNlcnQuY29tL3NoYTIt
# YXNzdXJlZC1jcy1nMS5jcmwwTAYDVR0gBEUwQzA3BglghkgBhv1sAwEwKjAoBggr
# BgEFBQcCARYcaHR0cHM6Ly93d3cuZGlnaWNlcnQuY29tL0NQUzAIBgZngQwBBAEw
# gYQGCCsGAQUFBwEBBHgwdjAkBggrBgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNl
# cnQuY29tME4GCCsGAQUFBzAChkJodHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20v
# RGlnaUNlcnRTSEEyQXNzdXJlZElEQ29kZVNpZ25pbmdDQS5jcnQwDAYDVR0TAQH/
# BAIwADANBgkqhkiG9w0BAQsFAAOCAQEAL3G7W2y46ZvBae5m6LR+rZu1yA1kkmk5
# 55rTY9lA9xCd0jrhY3cPzwKxGFy82surJeWDWbDIao7wqGXIfDrDgFahdM63btdl
# LWJacdfrM7l+82Zx22R30P/K5sVmk1UXr9eyOPzrYMiemUVru3xMNSZ4Md7lHrfb
# CODsEGjs+zWo4eqXWcnnOCxMd1cpr4PvKPonUhMAhCyFDD5TDzZDQIbHBDIvfb8j
# rg/cvZ2PArf5QZpLp5CiSGnGpC+O9rZhYbsOmcJVFMWMM0pMEw46T9HXHxHMfDC1
# oYTp8Xyc/JxiQ+084oNoGng4esFZIb18cJFXDfR0f4ie0pJTMvanqjCCBTAwggQY
# oAMCAQICEAQJGBtf1btmdVNDtW+VUAgwDQYJKoZIhvcNAQELBQAwZTELMAkGA1UE
# BhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2lj
# ZXJ0LmNvbTEkMCIGA1UEAxMbRGlnaUNlcnQgQXNzdXJlZCBJRCBSb290IENBMB4X
# DTEzMTAyMjEyMDAwMFoXDTI4MTAyMjEyMDAwMFowcjELMAkGA1UEBhMCVVMxFTAT
# BgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEx
# MC8GA1UEAxMoRGlnaUNlcnQgU0hBMiBBc3N1cmVkIElEIENvZGUgU2lnbmluZyBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAPjTsxx/DhGvZ3cH0wsx
# SRnP0PtFmbE620T1f+Wondsy13Hqdp0FLreP+pJDwKX5idQ3Gde2qvCchqXYJawO
# eSg6funRZ9PG+yknx9N7I5TkkSOWkHeC+aGEI2YSVDNQdLEoJrskacLCUvIUZ4qJ
# RdQtoaPpiCwgla4cSocI3wz14k1gGL6qxLKucDFmM3E+rHCiq85/6XzLkqHlOzEc
# z+ryCuRXu0q16XTmK/5sy350OTYNkO/ktU6kqepqCquE86xnTrXE94zRICUj6whk
# PlKWwfIPEvTFjg/BougsUfdzvL2FsWKDc0GCB+Q4i2pzINAPZHM8np+mM6n9Gd8l
# k9ECAwEAAaOCAc0wggHJMBIGA1UdEwEB/wQIMAYBAf8CAQAwDgYDVR0PAQH/BAQD
# AgGGMBMGA1UdJQQMMAoGCCsGAQUFBwMDMHkGCCsGAQUFBwEBBG0wazAkBggrBgEF
# BQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMEMGCCsGAQUFBzAChjdodHRw
# Oi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURSb290Q0Eu
# Y3J0MIGBBgNVHR8EejB4MDqgOKA2hjRodHRwOi8vY3JsNC5kaWdpY2VydC5jb20v
# RGlnaUNlcnRBc3N1cmVkSURSb290Q0EuY3JsMDqgOKA2hjRodHRwOi8vY3JsMy5k
# aWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURSb290Q0EuY3JsME8GA1UdIARI
# MEYwOAYKYIZIAYb9bAACBDAqMCgGCCsGAQUFBwIBFhxodHRwczovL3d3dy5kaWdp
# Y2VydC5jb20vQ1BTMAoGCGCGSAGG/WwDMB0GA1UdDgQWBBRaxLl7KgqjpepxA8Bg
# +S32ZXUOWDAfBgNVHSMEGDAWgBRF66Kv9JLLgjEtUYunpyGd823IDzANBgkqhkiG
# 9w0BAQsFAAOCAQEAPuwNWiSz8yLRFcgsfCUpdqgdXRwtOhrE7zBh134LYP3DPQ/E
# r4v97yrfIFU3sOH20ZJ1D1G0bqWOWuJeJIFOEKTuP3GOYw4TS63XX0R58zYUBor3
# nEZOXP+QsRsHDpEV+7qvtVHCjSSuJMbHJyqhKSgaOnEoAjwukaPAJRHinBRHoXpo
# aK+bp1wgXNlxsQyPu6j4xRJon89Ay0BEpRPw5mQMJQhCMrI2iiQC/i9yfhzXSUWW
# 6Fkd6fp0ZGuy62ZD2rOwjNXpDd32ASDOmTFjPQgaGLOBm0/GkxAG/AeB+ova+YJJ
# 92JuoVP6EpQYhS6SkepobEQysmah5xikmmRR7zCCBmowggVSoAMCAQICEAMBmgI6
# /1ixa9bV6uYX8GYwDQYJKoZIhvcNAQEFBQAwYjELMAkGA1UEBhMCVVMxFTATBgNV
# BAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8G
# A1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xMB4XDTE0MTAyMjAwMDAwMFoX
# DTI0MTAyMjAwMDAwMFowRzELMAkGA1UEBhMCVVMxETAPBgNVBAoTCERpZ2lDZXJ0
# MSUwIwYDVQQDExxEaWdpQ2VydCBUaW1lc3RhbXAgUmVzcG9uZGVyMIIBIjANBgkq
# hkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAo2Rd/Hyz4II14OD2xirmSXU7zG7gU6mf
# H2RZ5nxrf2uMnVX4kuOe1VpjWwJJUNmDzm9m7t3LhelfpfnUh3SIRDsZyeX1kZ/G
# FDmsJOqoSyyRicxeKPRktlC39RKzc5YKZ6O+YZ+u8/0SeHUOplsU/UUjjoZEVX0Y
# hgWMVYd5SEb3yg6Np95OX+Koti1ZAmGIYXIYaLm4fO7m5zQvMXeBMB+7NgGN7yfj
# 95rwTDFkjePr+hmHqH7P7IwMNlt6wXq4eMfJBi5GEMiN6ARg27xzdPpO2P6qQPGy
# znBGg+naQKFZOtkVCVeZVjCT88lhzNAIzGvsYkKRrALA76TwiRGPdwIDAQABo4ID
# NTCCAzEwDgYDVR0PAQH/BAQDAgeAMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAww
# CgYIKwYBBQUHAwgwggG/BgNVHSAEggG2MIIBsjCCAaEGCWCGSAGG/WwHATCCAZIw
# KAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMwggFkBggr
# BgEFBQcCAjCCAVYeggFSAEEAbgB5ACAAdQBzAGUAIABvAGYAIAB0AGgAaQBzACAA
# QwBlAHIAdABpAGYAaQBjAGEAdABlACAAYwBvAG4AcwB0AGkAdAB1AHQAZQBzACAA
# YQBjAGMAZQBwAHQAYQBuAGMAZQAgAG8AZgAgAHQAaABlACAARABpAGcAaQBDAGUA
# cgB0ACAAQwBQAC8AQwBQAFMAIABhAG4AZAAgAHQAaABlACAAUgBlAGwAeQBpAG4A
# ZwAgAFAAYQByAHQAeQAgAEEAZwByAGUAZQBtAGUAbgB0ACAAdwBoAGkAYwBoACAA
# bABpAG0AaQB0ACAAbABpAGEAYgBpAGwAaQB0AHkAIABhAG4AZAAgAGEAcgBlACAA
# aQBuAGMAbwByAHAAbwByAGEAdABlAGQAIABoAGUAcgBlAGkAbgAgAGIAeQAgAHIA
# ZQBmAGUAcgBlAG4AYwBlAC4wCwYJYIZIAYb9bAMVMB8GA1UdIwQYMBaAFBUAEisT
# mLKZB+0e36K+Vw0rZwLNMB0GA1UdDgQWBBRhWk0ktkkynUoqeRqDS/QeicHKfTB9
# BgNVHR8EdjB0MDigNqA0hjJodHRwOi8vY3JsMy5kaWdpY2VydC5jb20vRGlnaUNl
# cnRBc3N1cmVkSURDQS0xLmNybDA4oDagNIYyaHR0cDovL2NybDQuZGlnaWNlcnQu
# Y29tL0RpZ2lDZXJ0QXNzdXJlZElEQ0EtMS5jcmwwdwYIKwYBBQUHAQEEazBpMCQG
# CCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQQYIKwYBBQUHMAKG
# NWh0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRENB
# LTEuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCdJX4bM02yJoFcm4bOIyAPgIfliP//
# sdRqLDHtOhcZcRfNqRu8WhY5AJ3jbITkWkD73gYBjDf6m7GdJH7+IKRXrVu3mrBg
# JuppVyFdNC8fcbCDlBkFazWQEKB7l8f2P+fiEUGmvWLZ8Cc9OB0obzpSCfDscGLT
# Ykuw4HOmksDTjjHYL+NtFxMG7uQDthSr849Dp3GdId0UyhVdkkHa+Q+B0Zl0DSbE
# Dn8btfWg8cZ3BigV6diT5VUW8LsKqxzbXEgnZsijiwoc5ZXarsQuWaBh3drzbaJh
# 6YoLbewSGL33VVRAA5Ira8JRwgpIr7DUbuD0FAo6G+OPPcqvao173NhEMIIGzTCC
# BbWgAwIBAgIQBv35A5YDreoACus/J7u6GzANBgkqhkiG9w0BAQUFADBlMQswCQYD
# VQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGln
# aWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJvb3QgQ0Ew
# HhcNMDYxMTEwMDAwMDAwWhcNMjExMTEwMDAwMDAwWjBiMQswCQYDVQQGEwJVUzEV
# MBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29t
# MSEwHwYDVQQDExhEaWdpQ2VydCBBc3N1cmVkIElEIENBLTEwggEiMA0GCSqGSIb3
# DQEBAQUAA4IBDwAwggEKAoIBAQDogi2Z+crCQpWlgHNAcNKeVlRcqcTSQQaPyTP8
# TUWRXIGf7Syc+BZZ3561JBXCmLm0d0ncicQK2q/LXmvtrbBxMevPOkAMRk2T7It6
# NggDqww0/hhJgv7HxzFIgHweog+SDlDJxofrNj/YMMP/pvf7os1vcyP+rFYFkPAy
# IRaJxnCI+QWXfaPHQ90C6Ds97bFBo+0/vtuVSMTuHrPyvAwrmdDGXRJCgeGDboJz
# PyZLFJCuWWYKxI2+0s4Grq2Eb0iEm09AufFM8q+Y+/bOQF1c9qjxL6/siSLyaxhl
# scFzrdfx2M8eCnRcQrhofrfVdwonVnwPYqQ/MhRglf0HBKIJAgMBAAGjggN6MIID
# djAOBgNVHQ8BAf8EBAMCAYYwOwYDVR0lBDQwMgYIKwYBBQUHAwEGCCsGAQUFBwMC
# BggrBgEFBQcDAwYIKwYBBQUHAwQGCCsGAQUFBwMIMIIB0gYDVR0gBIIByTCCAcUw
# ggG0BgpghkgBhv1sAAEEMIIBpDA6BggrBgEFBQcCARYuaHR0cDovL3d3dy5kaWdp
# Y2VydC5jb20vc3NsLWNwcy1yZXBvc2l0b3J5Lmh0bTCCAWQGCCsGAQUFBwICMIIB
# Vh6CAVIAQQBuAHkAIAB1AHMAZQAgAG8AZgAgAHQAaABpAHMAIABDAGUAcgB0AGkA
# ZgBpAGMAYQB0AGUAIABjAG8AbgBzAHQAaQB0AHUAdABlAHMAIABhAGMAYwBlAHAA
# dABhAG4AYwBlACAAbwBmACAAdABoAGUAIABEAGkAZwBpAEMAZQByAHQAIABDAFAA
# LwBDAFAAUwAgAGEAbgBkACAAdABoAGUAIABSAGUAbAB5AGkAbgBnACAAUABhAHIA
# dAB5ACAAQQBnAHIAZQBlAG0AZQBuAHQAIAB3AGgAaQBjAGgAIABsAGkAbQBpAHQA
# IABsAGkAYQBiAGkAbABpAHQAeQAgAGEAbgBkACAAYQByAGUAIABpAG4AYwBvAHIA
# cABvAHIAYQB0AGUAZAAgAGgAZQByAGUAaQBuACAAYgB5ACAAcgBlAGYAZQByAGUA
# bgBjAGUALjALBglghkgBhv1sAxUwEgYDVR0TAQH/BAgwBgEB/wIBADB5BggrBgEF
# BQcBAQRtMGswJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBD
# BggrBgEFBQcwAoY3aHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0
# QXNzdXJlZElEUm9vdENBLmNydDCBgQYDVR0fBHoweDA6oDigNoY0aHR0cDovL2Ny
# bDMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9vdENBLmNybDA6oDig
# NoY0aHR0cDovL2NybDQuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9v
# dENBLmNybDAdBgNVHQ4EFgQUFQASKxOYspkH7R7for5XDStnAs0wHwYDVR0jBBgw
# FoAUReuir/SSy4IxLVGLp6chnfNtyA8wDQYJKoZIhvcNAQEFBQADggEBAEZQPsm3
# KCSnOB22WymvUs9S6TFHq1Zce9UNC0Gz7+x1H3Q48rJcYaKclcNQ5IK5I9G6OoZy
# rTh4rHVdFxc0ckeFlFbR67s2hHfMJKXzBBlVqefj56tizfuLLZDCwNK1lL1eT7EF
# 0g49GqkUW6aGMWKoqDPkmzmnxPXOHXh2lCVz5Cqrz5x2S+1fwksW5EtwTACJHvzF
# ebxMElf+X+EevAJdqP77BzhPDcZdkbkPZ0XN1oPt55INjbFpjE/7WeAjD9KqrgB8
# 7pxCDs+R1ye3Fu4Pw718CqDuLAhVhSK46xgaTfwqIa1JMYNHlXdx3LEbS0scEJx3
# FMGdTy9alQgpECYxggQ7MIIENwIBATCBhjByMQswCQYDVQQGEwJVUzEVMBMGA1UE
# ChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYD
# VQQDEyhEaWdpQ2VydCBTSEEyIEFzc3VyZWQgSUQgQ29kZSBTaWduaW5nIENBAhAD
# pwWPYXZWpp6sK+IYC72NMAkGBSsOAwIaBQCgeDAYBgorBgEEAYI3AgEMMQowCKAC
# gAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsx
# DjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBRaxctLVRcagWgS/y3Q43VK
# 9Rx4lDANBgkqhkiG9w0BAQEFAASCAQA4bxGeSVAgkcItoi3RzYVdc0IapsZkuKu3
# 8Y/ZMqQSff7EvgBP/kW4i6DudKpd8R9XKnrmeo95D7OtY3gca0A5m6GbZoFO40dT
# RrJ2XJr+09f3xD1gIRfVdutmj7603qhJrZfditqB0Vo7EV2K1BOa90PKStHFJicj
# 9nVJ15FxGQeNjFVvIVnftzn7qjNCD5zplXLpm3kH1y/qPvdNM8mg7SpzTHbK4kqh
# mS2HmSOq9wjjfpvyiSe8OaTczjSVGQDV7j/Vjad8xMgBWEVL1ZXjmfl1qdhmWdnc
# 2aqLQFTINXcK0+PPW8DwADOPNWxk8oKFZei47qT8aDOFhtiB0N7BoYICDzCCAgsG
# CSqGSIb3DQEJBjGCAfwwggH4AgEBMHYwYjELMAkGA1UEBhMCVVMxFTATBgNVBAoT
# DERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEhMB8GA1UE
# AxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xAhADAZoCOv9YsWvW1ermF/BmMAkG
# BSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJ
# BTEPFw0xNzAyMDgwNzA0MjVaMCMGCSqGSIb3DQEJBDEWBBQmL45oqBbtYYKzVEs4
# 8K2/QOkRJjANBgkqhkiG9w0BAQEFAASCAQAHebAKT5Ss63RKlkOe0hUyHoLW0aXr
# QpqSxP7UJtxOxDI/ArfzAH9baNdbBIEzB0zt9yGLAQnu1qBEjkVKJYEsceHZfjwk
# P+2z9bYE9W82iNIVe8iMqUsLcSLRXkn+3o4bhZwM4yQ16+ZDkvDQl5X/gLZ+7NCM
# uRvzlYIjBVo6ycqKKxcu1d4/G2shaOZzYsv/vQx002ht6AdS/PHiC9Pbi7O1iVlh
# PbXL+fOPQ8/zfNZMvFw9Sx7CZtsIhO7wvZJKnavMCDhxCEmntvX3vY6F0QW6jh4U
# 5EStH3kgn1sHFMU57oe8Kp0Wkz/WBJp/dapfzq38q4eQxZrNzTnOupVJ
# SIG # End signature block
